namespace mylib
{
	using System.Drawing;
	using System.Drawing.Drawing2D;
	using System.Collections;

	using System;
	using mylib;

#if MDX
	using System.Drawing.Imaging;
	using Microsoft.DirectX;
	using Microsoft.DirectX.Direct3D;
	using DirectX=Microsoft.DirectX;
	using Direct3D=Microsoft.DirectX.Direct3D;
#endif


	public interface ConsoleDebugger{
			void DBG(string str);
			void DBGn(string str);
	}

	public interface ScreenManager : ConsoleDebugger{
//			bool Drawing{get;}
			double DrawX{get;}
			double DrawY{get;}
			double VirtualWidth{get;}
			double VirtualHeight{get;}
			double ScreenWidth{get;}
			double ScreenHeight{get;}
			double ScrollTop{get;set;}
			double ScrollLeft{get;set;}
			double StageWidth{get;}
			double StageHeight{get;}
			double ScrollXmax{get;}
			double ScrollYmax{get;}
			double ScrollXmin{get;}
			double ScrollYmin{get;}
			bool   ScrollLock{get;set;}
			double BlockWidth{get;}
			double BlockHeight{get;}
			int  Mozaic{get;set;}
			bool ClsFlag{get;}
			Image GetImage(string str);
			int GetImageID(string str);
			Image GetImageData(int n);
			string GetImageName(int n);
			bool DrawEntry(Sprite s);
			
			int CheckScreen(MCharacter m);
			int CheckScreen(Sprite s);
			int CheckScreen(double x,double y,double w,double h);

			string CodeBase{get;set;}

#if MDX
			Texture GetTextureData(int n);
			Device DxDevice {get;set;}
			Direct3D.Sprite DxSprite {get;set;}
#endif
	}
	
	public enum WriteMode{Vram=1,Ram=2,All=3}

	public interface StageManager : ScreenManager{
			int GameTimer{get;set;}

			int GetXYPointObject(double x,double y);
			int GetXYBlockObject(int x,int y);
			int SetXYPointObject(double x,double y,int code,WriteMode flag);
			int SetXYBlockObject(int x,int y,int code,WriteMode flag);
			int SetXYPointObject(double x,double y,int code);
			int SetXYBlockObject(int x,int y,int code);
			
			int GetWorldNumber();
			int GetAreaNumber(); 
			int GetStageMode(); 
			
			int CheckWallBlock(int n);
			int CheckFloorBlock(int n);
			int CheckUpperBlock(int n);
			int CheckCoin(int n);
			int CheckEventBlock(int n);
			int CheckWaterStream(int n);

			Player GetPlayerObject();
			Player GetPlayerObject(int n);
//			Player SetPlayerObject(Player p);

			string GetBlockImageName(int n);

			MCharacter GetBoss();
			void SetBoss(MCharacter m);
			void KillBoss();
			MCharacter GetGoldenAxe();
			void SetGoldenAxe(MCharacter m);
			void KillGoldenAxe();
			MCharacter GetGoalFlagObject();
			void SetGoalFlag(double xx,double yy,bool v);
			MCharacter GetCastleFlagObject();
			void SetCastleFlag(double xx,double yy,bool v);
			void SetPrincess(MCharacter m);
			MCharacter GetPrincess();
			void RescuePrincess();


			MCharacter CreateXYBlockObject(int code,double x,double y,int option);
			MCharacter CreateObject(int code,double x,double y,int option);
			MCharacter CreateObject(int code,double x,double y,int option,ObjGroup group);
			MCharacter CreateXYBlockEnemy(int code,double x,double y,int option);
			MCharacter CreateEnemy(int code,double x,double y,int option);
			MCharacter CreateEnemy(int code,double x,double y,int option,ObjGroup group);
			MCharacter CreateEffect(int code,double x,double y,int option);
			MCharacter CreateEffect(int code,double x,double y,int option,ObjGroup group);

			MCharacter CreateMessage(string str,double x,double y);
			
			MCharacter CenterMessage(string str,double y);
			MCharacter CenterMessage(string str,double y,int showtime);
			MCharacter CenterMessage(string str,double y,int showtime,System.Drawing.Font font);

			void CallEnemy();
//			MCharacter CallEnemy();
			MCharacter CallEnemy(int code);

			ObjGroup Enemies{get; }
			ObjGroup Others {get; }
		 	ObjGroup Effects{get; }

			GameStage GameStage{get;}

			int M(int x);
			int M(double x);

	}

	public interface GameEventManager{
		int EventDrop(Player p);
		int EventDie(Player p);
		int EventWarp(Player p);
		int EventWarp(Player p,int id);
		int EventEventBlock(Player p,int block);
		int EventGoalBlock(Player p,int block);
		int EventUpper(Player p,int block,double x,double y);
		int EventPrincess(Player p);

	}

	
	
	class GameException:System.Exception
	{
		public const int StageClear = 1;
		public const int Goal = 2;
		public const int Die = 3;
		public const int Warp = 4;

		public int Code1 = 0;
		public int Code2 = 0;
		public GameException(int msg1, int msg2)
		{
			Code1 = msg1;
			Code2 = msg2;
		}
	}
	//


	internal interface GameEventExceptionListner
		{
/*
			bool eventStageInit(EnvManager E, Player p);
			bool eventScroll(EnvManager E, Player p);
			bool eventGoal(EnvManager E, Player p);
			bool eventWarpBlock(EnvManager E, Player p);
			bool eventEventBlock(EnvManager E, Player p);
			bool eventDie(EnvManager E, Player p);
			bool eventDrop(EnvManager E, Player p);
			bool eventTimeover(EnvManager E, Player p);
			bool eventStageClear(EnvManager E, Player p);
*/
		}
	


	public class EnvManager : StageManager,GameEventManager{
		private bool drawing=false;
		 public bool Drawing{
			get{return drawing;}
		 }


		private double screenWidth  = 512;
		private double screenHeight = 448;
		private double virtualWidth  = 512;
		private double virtualHeight = 448;
		private double scrollTop =0;
		private double scrollLeft=0;
		private double scrollXmax =640;
		private double scrollYmax =480;
		private double scrollXmin =0;
		private double scrollYmin =0;
		private double autoScrollX;
		private double autoScrollY;

		 public double ScreenWidth{
			get{
				return screenWidth;
			}
		 }

		 public double ScreenHeight{
			get{
				return screenHeight;
			}
		 }
		 public double VirtualWidth{
			get{
				return virtualWidth;
			}
		 }

		 public double VirtualHeight{
			get{
				return virtualHeight;
			}
		 }
		 public double ScrollTop{
			get{
				return scrollTop;
			}
			set{
				scrollTop=value;
				if(scrollTop>scrollYmax)scrollTop=scrollYmax;
				if(scrollTop<scrollYmin)scrollTop=scrollYmin;
			}
		 }

		 public double ScrollLeft{
			get{
				return scrollLeft;
			}
			set{
				scrollLeft=value;
//				if(scrollLeft>scrollXmax)scrollLeft=scrollXmax;
//				if(scrollLeft<scrollXmin)scrollLeft=scrollXmin;
			}
		 }

		 public double ScrollXmax{
			get{
				return scrollXmax;
			}
		 }
		 public double ScrollYmax{
			get{
				return scrollYmax;
			}
		 }
		 public double ScrollXmin{
			get{
				return scrollXmin;
			}
		 }
		 public double ScrollYmin{
			get{
				return scrollYmin;
			}
		 }
		 public double StageWidth{
			get{
				return scrollXmax+virtualWidth;
			}
		 }
		 public double StageHeight{
			get{
				return scrollYmax+virtualHeight;
			}
		 }



		private const int stageZPosition=100;


		private bool scrollLock=false;
		 public bool ScrollLock{
			get{
				return scrollLock;
			}
			set{
				scrollLock=value;
			}
		 }

		private double blockWidth =32;
		private double blockHeight=32;

		public double BlockWidth{
			get{
				return blockWidth;
			}
			set{
				blockWidth=value;
			}
		}

		public double BlockHeight{
			get{
				return blockHeight;
			}
			set{
				blockHeight=value;
			}
		}

		private int mozaic=0;
		private bool clsFlag=false;
		
		 public int Mozaic{
			get{
				return mozaic;
			}
			set{
				mozaic=value;
			}
		 }
		 
		 public bool ClsFlag{
			get{	return clsFlag;}
			set{	clsFlag=value;}
		 }


			public const int Gaming = 0;
			public const int Waiting= 1;
			public const int Title  = 2;
			public const int Ending = 3;
		private int mode=Title;
		 public int Mode{
			get{return mode;}
			set{mode=value;}
		 }
		private int gameClock=0;
		private int gameStartClock=0;
//		private int waitTimer=0;

		private static int timerDefault=300;
		private int gameTimer=timerDefault;
		 public int GameTimer{
			get{
				return gameTimer;
			}
			set{
				gameTimer=value;
			}
		 }

		private bool pause=false;


		
		private int currentStage=0;
		private int stageMode=0;//StageMode.NORMAL;
		 public int GetStageMode(){
			return stageMode;
		 }

		private bool uraMode=false;
		private bool retryFlag=false;
		private bool warpFlag=false;
		private int warpID=-1;

		private ArrayList lastStage=new ArrayList();
		private int hiScore=GameConfig.HiScore;
		 public int HiScore{
			get{return hiScore;}
		 }
		private int lastWorld=1;


		private int callEnemyRate=100;
		private int callEnemyCnt=100;

		private double wx=1;
		private double wy=1;

		 public double DrawX{
			get{	return wx;}
		 }
		 public double DrawY{
			get{	return wy;}
		}

		private mylib.Random R;
		
		private GameStage gameStage;
		 public GameStage GameStage{
			get {return gameStage;}
		 }
		private GamePad gamePad1;


		public ObjGroup enemies = new ObjGroup();
		public ObjGroup others  = new ObjGroup();
		public ObjGroup effects = new ObjGroup();
		 public ObjGroup Enemies{get{return enemies;} }
		 public ObjGroup Others {get{return others; } }
		 public ObjGroup Effects{get{return effects;} }


		private Image offScreen;
		private Graphics graphics;
		private Brush clsBrush = Brushes.Black;
		private PixelEffector pixelEffector;

		private ILoader iLoader;
		private string codeBase="";//=System.AppDomain.CurrentDomain.BaseDirectory;
		 public string CodeBase{
			get{	return codeBase;}
			set{	codeBase=value;	
			}
		 }
		private ZSorter zSorter;
		private WaitScreen wScreen;
		private TitleScreen tScreen;
		
		private System.Drawing.Font gameFont;//  = new Font();
		private System.Drawing.Font scoreFont;// = new Font();
		private System.Drawing.Font infoFont;//  = new Font();
		 public System.Drawing.Font GameFont{
			get{	return gameFont;}
			set{	gameFont=value;}
		 }
		 public System.Drawing.Font ScoreFont{
			get{	return scoreFont;}
			set{	gameFont=value;}
		 }
		 public System.Drawing.Font InfoFont{
			get{	return infoFont;}
			set{	infoFont=value;}
		 }
		 public Color GameFontColor{
			get{	return gameFontColor;}
			set{
				gameFontColor=value;
				gameBrush=new SolidBrush(value);
				}
		 }
		 public Color ScoreFontColor{
			get{	return scoreFontColor;}
			set{
				scoreFontColor=value;
				scoreBrush=new SolidBrush(value);
				}
		 }
		 public Color InfoFontColor{
			get{	return infoFontColor;}
			set{
				infoFontColor=value;
				infoBrush=new SolidBrush(value);
				}
		 }
		private Brush gameBrush  = Brushes.White;
		private Brush scoreBrush = Brushes.White;
		private Brush infoBrush  = Brushes.White;
		private Color gameFontColor  = Color.White;
		private Color scoreFontColor = Color.White;
		private Color infoFontColor  = Color.White;



		private ScoreBoard scoreBoard;

		
		private DummyObject dummyObject;// =new DummyObject(this,0,0);
		private GoalFlag   goalFlag;// =new GoalFlag(this,0,0,0);
		private CastleFlag castleFlag ;//new CastleFlag(0,0,this,0);
		private MCharacter boss	;//	= dummyObject;
		private MCharacter goldenAxe;//	= dummyObject;
		private MCharacter princess;

		private Player maryo;

#if MDX
		private Device device;
		private Device currentDevice;
		private Microsoft.DirectX.Direct3D.Sprite dxSprite;

		private Texture offTexture;
		private Surface offSurface;
		private RenderToSurface render;
		private Microsoft.DirectX.Matrix scaleMatrix = new Microsoft.DirectX.Matrix();
		private Rectangle screenRect = new Rectangle(0,0,0,0);
		private Vector3 screenPos   =new Vector3(0,0,0);
		private Vector3 screenCenter=new Vector3(0,0,0);

		public Device DxDevice{
//			get{return currentDevice;}
			get{return device;}
			set{
				device=value;
				if(currentDevice==null)currentDevice=device;
			}
		}
		public Direct3D.Sprite DxSprite{
			get{return dxSprite;}
			set{dxSprite=value;}
		}
		public EnvManager(string cb,int w,int h,Device dev){
			CodeBase=cb;
			device=dev;
			initGraphics(w,h);

			Sprite.Manager=this;

			R=new mylib.Random();
			gamePad1=new GamePad();
			iLoader=new ILoader(this);
			
			scoreBoard=new ScoreBoard(this);


			maryo = new Player(this,this,0,0,gamePad1);

			dummyObject=new DummyObject(this,0,0);

			goalFlag   =new GoalFlag(this,0,0,0,0);
			castleFlag =new CastleFlag(this,0,0,0,0);

			zSorter   = new ZSorter();
			gameStage = new GameStage(this);
			wScreen   = new WaitScreen(this);
			tScreen   = new TitleScreen(this,gamePad1);

		}

#endif


	
//-----------------------------------------
		public EnvManager(string cb,int w,int h){
			CodeBase=cb;
			initGraphics(w,h);

			Sprite.Manager=this;

			R=new mylib.Random();
			gamePad1=new GamePad();
			iLoader=new ILoader(this);
			
			scoreBoard=new ScoreBoard(this);


			maryo = new Player(this,this,0,0,gamePad1);

			dummyObject=new DummyObject(this,0,0);

			goalFlag   =new GoalFlag(this,0,0,0,0);
			castleFlag =new CastleFlag(this,0,0,0,0);

			zSorter   = new ZSorter();
			gameStage = new GameStage(this);
			wScreen   = new WaitScreen(this);
			tScreen   = new TitleScreen(this,gamePad1);


		}
		private void initGraphics(int w,int h){
			screenWidth=w;screenHeight=h;

			offScreen = new Bitmap(w*2,h*2);
//			((Bitmap)offScreen).MakeTransparent(Color.White);
			graphics = Graphics.FromImage(offScreen);
//			graphics.InterpolationMode = InterpolationMode.NearestNeighbor;
			graphics.SmoothingMode = SmoothingMode.None;
			graphics.TranslateTransform(w/2, h/2);

			pixelEffector = new PixelEffector(this,offScreen);
#if MDX
			initMDX(w,h);
#endif

		}
		
		public void Start(){
			scoreBoard.Init(graphics,scoreFont,scoreBrush);
			wScreen.SetLogoMode(graphics,150);
		}

		public void SetGamePad(GamePad p){
			gamePad1=p;
		}
		public GamePad GetGamePad(){
			return gamePad1;
		}
		public Graphics GetGraphics(){
			return graphics;
		}
		public System.Drawing.Font SetGameFont(System.Drawing.Font m,Color c){
			gameFont=m ; GameFontColor  =c;
			Sprite.DefaultFont=gameFont;
			Sprite.DefaultColor=GameFontColor;
			return gameFont;
		}
		public System.Drawing.Font SetScoreFont(System.Drawing.Font m,Color c){
			scoreFont=m;ScoreFontColor =c;
			return gameFont;
		}
		public System.Drawing.Font SetInfoFont(System.Drawing.Font m,Color c){
			infoFont=m ; InfoFontColor  =c;
			return gameFont;
		}

		public void SetVirtualScreenSize(double w,double h){
			virtualWidth=w;
			virtualHeight=h;
			setWxWy();
		}
		private void setWxWy(){
			wx=screenWidth/virtualWidth;
			wy=screenHeight/virtualHeight;
			scrollXmax=gameStage.Cols*blockWidth-virtualWidth;
			scrollYmax=gameStage.Rows*blockHeight-virtualHeight;
		}
//-----------------------------------------
		public int M(int x){return (R.M % x);}
		public int M(double x){return M((int)Math.Round(x));}
//-----------------------------------------

		public void GameStart(){

		  try{
			GameReset();
			 maryo.Reset(1);
			 gameStartClock=gameClock;
			lastStage.Clear();

			StageChange(maryo,currentStage);
			
			
			wScreen.SetInfoMode(graphics,90);
		   }catch(Exception e){
			DBG("Exception "+ e + "in EnvManager.GameStart");
		   }

		}
		public void GameReset(){
			stageMode=StageMode.Normal;
			scrollTop=0;
			scrollLeft=0;
			scrollLock=false;
			currentStage=0;

			warpFlag  = false;
			retryFlag = false;
			maryo.Reset();
			Enemies.Clear();
			Others.Clear();
			Effects.Clear();

			maryo.SetEnemyGroup(Enemies);

			System.GC.Collect();
		}


		public string StageReset(Player p){
			scrollTop=0;
			scrollLeft=0;
			scrollLock=false;
			Enemies.Clear();
			Others.Clear();
			Effects.Clear();
			if(currentStage > StageMode.StageSrc.Length)
				currentStage=StageMode.StageSrc.Length-1;
			StageChange(p,StageMode.StageSrc[currentStage]);
			System.GC.Collect();


			return StageMode.StageSrc[currentStage];
		} //

		public string BackToLastStage(Player p,bool f){
			if(f)warpFlag=true;
			
			if(lastStage.Count<2){return SetWorldNumber(p,0);}
			 string str=(string)lastStage[(lastStage.Count-2)];
			StageChange(p,str);
			return str;
		} //

		public int NextStage(Player p){
		 scrollTop=0;
		 scrollLeft=0;
		 scrollLock=false;
		 retryFlag=false;
		 warpFlag=false;
		 warpID=-1;
		 currentStage++;
		  if(currentStage < StageMode.StageSrc.Length)
					StageChange(p,StageMode.StageSrc[currentStage]);
		  else {
			uraMode=true;
			currentStage--;
			Gameover(p,1);
		  }

		 return currentStage;
		} //

		public bool StageChange(Player p,int src){
			return StageChange(p,StageMode.StageSrc[src]);
		}
		public bool StageChange(Player p,String src){

			lastStage.Add(src);
			LoadStage(src);

			scrollTop=0;
			scrollLeft=0;
			scrollLock=false;

			EnemyParameter.LiveCannonCnt=0;
			EnemyParameter.LiveCnt=0;

			p.Reset(0);
			p.StopAnimate();
			 if(stageMode==StageMode.Water)p.ActionMode=ActionMode.Swim;
			 else p.ActionMode=ActionMode.Run;

			DBGn(" warpID="+warpID+" warpFlag="+warpFlag);
			 if(!warpFlag && gameStage.GameTimer>=0)gameTimer=gameStage.GameTimer;

			 if(warpFlag && warpID>=0){
				p.X=gameStage.WarpStartX[warpID];
				p.Y=gameStage.WarpStartY[warpID];
				p.HFace=gameStage.WarpStartHFace[warpID];
				Scroll();
				p.SetAnimateMode(gameStage.WarpStartAnime[warpID]);
			 }else
			 if(retryFlag){
				p.X=gameStage.PLRestartX;
				p.Y=gameStage.PLRestartY;
				p.HFace=gameStage.PLRestartHFace;
				Scroll();
				p.SetAnimateMode(gameStage.PLRestartAnime);
			 }else{
				p.X=gameStage.PLStartX;
				p.Y=gameStage.PLStartY;
				p.HFace=gameStage.PLStartHFace;
				Scroll();
				p.SetAnimateMode(gameStage.PLStartAnime);
			}

			scrollLeft = p.X-virtualWidth/2;
			scrollTop  = p.Y-virtualHeight*.7;
			if(scrollLeft<scrollXmin)scrollLeft=scrollXmin;
			if(scrollTop <scrollYmin)scrollTop=scrollYmin;


			p.Y-=p.SpriteHeight;
			retryFlag=false;
			warpFlag=false;



			CreateStageObjects();
			showInfoText();
			DBGn("currentStage="+currentStage);
			if(currentStage==0){
				DBGn("stageZeroInit");
				stageZeroInit();
			}

			wScreen.SetInfoMode(graphics,90);
			System.GC.Collect();
		 return true;
		}
		public bool LoadStage(string src){
			gameStage.Load(src);

				// setScreenSizeData(screenWidth,screenHeight,gameStage.virtualWidth,gameStage.virtualHeight);
			SetVirtualScreenSize(gameStage.VirtualWidth,gameStage.VirtualHeight);

			scrollXmax=gameStage.Cols*blockWidth -virtualWidth;
			scrollYmax=gameStage.Rows*blockHeight-virtualHeight;
			 stageMode=gameStage.StageMode;
			 EnemyParameter.Lift_liftWidth=gameStage.LiftWidth;
			 callEnemyCnt=gameStage.EnemyRate;
			 EnemyParameter.LiveCannonMax = gameStage.MaxEnemies;
			 autoScrollX=gameStage.AutoScrollX;
			 autoScrollY=gameStage.AutoScrollY;
			WarpHall.Reset();
			goalFlag.Reset();
			 return true;
		}
		public bool CreateStageObjects(){
			enemies.Clear();
			others.Clear();
			effects.Clear();
			gameStage.CreateObjects();
			return true;
		}

		public void Gameover(Player p,int flag){
			 if(hiScore<p.Score)hiScore=p.Score;
			 if(lastWorld<GetWorldNumber())lastWorld=GetWorldNumber();
			 DBGn("Game over!");
			wScreen.SetGameoverMode(graphics,30*3,flag);
			 p.Reset(1);
		}

		public void stageZeroInit(){
			MCharacter logo= new MCharacter(this,0,0);
			logo.GroupEntry(Others);
			logo.Sprite.BitmapImageName="img/logo.gif";

			logo.SetSize(384,150,384,150);
			logo.Visible=true;
			logo.MoveTo(64,0,Zindex.PlayerZ+1);
			logo.XMove  =   0;
			logo.YMove  = -1;
			logo.YAccel = 0.1;
			logo.YJump  = 6;
			logo.Sprite.Opacity=.7;

			MCharacter go=CenterMessage("GO!",blockHeight*10,10000);
			go.Sprite.Color=Color.Yellow;
			go.X=blockWidth*12;
			go.YMove=0;


#if mytest
#else
			// ReBj[\[h܂łXN[ł悤ɂ
			double tmp=scrollXmax;
			scrollXmax = blockWidth * 2 * (lastWorld -1);
			if(scrollXmax>tmp)scrollXmax=tmp;
#endif
		}

	//-----------------------------
		public int EventDie(Player p){
			gameTimer=GameConfig.DefaultGameTimer;
			warpFlag=false;
			warpID=-1;
			
			 // SrÅȈՑΉ
			 // cXN[ʁEʂ͖Ή
			  int plx1=gameStage.PLStartX;
			  int plx2=gameStage.PLRestartX;
			 if(plx2>0 && plx1!=plx2){
			  if( ((p.X > plx2) && (plx1 < plx2)) ||	// EXN[
			      ((p.X < plx2) && (plx1 > plx2))		// XN[
			      )retryFlag=true;
			 }
			 if(p.Life > 0)
				StageChange(p,StageMode.StageSrc[currentStage]);
			 else
				Gameover(p,0);
			 return -1;
		}
		public int EventDrop(Player p){
			if(stageMode==StageMode.Heaven){
				throw new GameException(GameException.Warp,-1);
			//	return 1;
			}
			return -1;
		}
		public int EventWarp(Player p){return EventWarp(p,warpID);}
		public int EventWarp(Player p,int id){
//			DBGn("EventWarp!");
//			DBGn("EventWarp!"+gameStage.WarpSrc[0]);
			if((id < 0) ||
			   (gameStage.WarpSrc[id]==null) ||
			    gameStage.WarpSrc[id].Equals("") ||
			    gameStage.WarpSrc[id].Equals("%RETURN%")
				){
			DBG("backto laststage.Count="+(lastStage.Count)+" "+
			(string)lastStage[(lastStage.Count-1)]+"\n");
				DBGn("WARP=>RETURN");
				DBGn(BackToLastStage(p,true));
				//warpID=-1;
				return 1;
			}else
			if(gameStage.WarpSrc[id].Equals("%CLEAR%")){
				DBGn("WARP=>STAGE CLEAR");
				warpID=-1;
				NextStage(p);
				return 2;
			}else
			if(gameStage.WarpSrc[id].IndexOf("WORLD=")>=0){
				int c=gameStage.WarpSrc[id].IndexOf("=");
	//			DBG("trim="+gameStage.warpSrc[id].substring(c+1).trim());
				int world=System.Int32.Parse(gameStage.WarpSrc[id].Substring(c+1).Trim());

				DBGn("WARP=>WORLD "+world);
				SetWorldNumber(p,world);
				warpID=-1;
				return 3;
			
			}else{
				warpFlag=false;
				DBGn("WARP=>STAGE "+gameStage.WarpSrc[id]);
				warpID=id;
				StageChange(p,gameStage.WarpSrc[id]);
				return 4;
			}
		}
		public int EventEventBlock(Player p,int block){
			if(block>=0x6d && block<=0x6e){
				p.GoalFlag();
				return 1;
			}
			return -1;
		}
		public int EventGoalBlock(Player p,int block){
			NextStage(p);
			return 1;
		}

	
		public int EventUpper(Player p,int block,double x,double y){
			MCharacter m;
			 int xx=(int)Math.Floor(((int)Math.Floor(x / blockWidth )) * blockWidth);
			 int yy=(int)Math.Floor(((int)Math.Floor(y / blockHeight)) * blockHeight);
			 if( 
			 	(block >=0x10 &&block <=0x3f) ||
			 	(block == Blocks.SoftBlock) || (block == Blocks.Ground)
			   ){
				 if( CheckCoin( GetXYPointObject(x,y-blockHeight) ) > 0 ){
					CreateEffect(CharCode.BounceCoinEffect,x,y-blockHeight,0);
					p.GetItem(ItemCode.Coin);
					SetXYPointObject(x,y-blockHeight,0);
				 }
			 }

			if(block == Blocks.SoftBlock || block == Blocks.Ground){
				 m=CreateObject(CharCode.Bounce,x,y,block);
				 m.SetMaster(p);
				 if(p.Mode != PlayerMode.Normal){
					SetXYPointObject(x,y,0);
					p.GetScore(50);
					m.Visible=false;
					CreateEffect(CharCode.BrokenEffect,x,y,block);
				 }
				return 1;
			}
			
			if( block >= 0x10 && block <= 0x3f ){
			  switch(block % 0x10){

				case 0x0:
				 m=CreateObject(CharCode.Bounce,x,y,block);
				 m.SetMaster(p);
				 SetXYPointObject(x,y,3);
				 CreateEffect(CharCode.BounceCoinEffect,x,y,0);
					p.GetItem(ItemCode.Coin);
				return 1;

				case 0x1:
				 m=CreateObject(CharCode.Bounce,x,y,block);
				 m.SetMaster(p);
				 SetXYPointObject(x,y,3);
				 if(p.Mode==PlayerMode.Normal)
					CreateObject(CharCode.Kinoko,x,y,0);
				 else	CreateObject(CharCode.Flower,x,y,0);
				return 1;

				case 0x2:
				 m=CreateObject(CharCode.Bounce,x,y,block);
				 m.SetMaster(p);
				 SetXYPointObject(x,y,3);
					CreateObject(CharCode.Star,x,y,0);
				return 1;

				case 0x3:
				 m=CreateObject(CharCode.Bounce,x,y,block);
				 m.SetMaster(p);
				 SetXYPointObject(x,y,3);
					CreateObject(CharCode.Iget,x,y,0);
				return 1;

				case 0x4:
				 m=CreateObject(CharCode.Bounce,x,y,block);
				 m.SetMaster(p);
				 SetXYPointObject(x,y,3);
					CreateObject(CharCode.Tsuta,x,y,0);
				return 1;

				case 0x5:
				 m=CreateObject(CharCode.TenCoin,x,y,block);
				 m.SetMaster(p);

				 m=CreateObject(CharCode.Bounce,x,y,block);
				 m.SetMaster(p);
				 SetXYPointObject(x,y,5,WriteMode.Vram);
				 CreateEffect(CharCode.BounceCoinEffect,x,y,0);
					p.GetItem(ItemCode.Coin);
				return 1;

			  }
			}
			return -1;
		}
	
		public int EventPrincess(Player p){
			Princess pr=(Princess)GetPrincess();
			if(pr==null)return -1;
			pr.ShowMessage();
			return 1;
		}


	//-----------------------------
		public int GetWorldNumber(){
			int vc=currentStage-1;
			if(vc<0)return 0;
			return (int)Math.Floor(vc/4)+1;
		} //

		public int GetAreaNumber(){
			int vc=currentStage-1;
			if(vc<0)return 0;
			return ((vc%4)+1);
		} //

		public string SetWorldNumber(Player p,int n){ //WARP
			int w=n*4;
			if(w>StageMode.StageSrc.Length+3)w=0;
			currentStage=w-3;
			 if(currentStage<0)currentStage=0;
			retryFlag=false;
			warpFlag=false;
//---------
			StageChange(p,StageMode.StageSrc[currentStage]);
			return "";
		} //


		//-----------------------------
			public void Pause(){Pause(!pause);}
			public void Pause(bool b){
				if(mode!=Gaming)return;
				pause=b;
				 if(pause){
					MCharacter m=CenterMessage("P A U S E",virtualHeight*.5,100,gameFont);
					m.Sprite.Color=Color.Red;
					m.Z=1000;
					m.XMove= M(20)-10;
					m.YMove=-M(30);
					m.YAccel=1;
					//DBG(""+M(30));
				 }
			}



		public Image GetImage(string str){
			return iLoader.GetImage(str);
		}
		public int GetImageID(string str){
			return iLoader.GetImageID(str);
		}
		public Image GetImageData(int n){
			return iLoader.GetImageData(n);
		}
		public string GetImageName(int n){
			return iLoader.GetImageName(n);
		}


		public void Draw(){
		 drawing=true;
		 try{
		  if(mode==Gaming || mode==Title){
			maryo.DrawEntry();
			Enemies.DrawEntry();
			Others.DrawEntry();
			Effects.DrawEntry();
#if MDX
			if(device==null)return;
			if(device.Disposed)return;

			if(offTexture.Disposed)ResetDevice();

			Sprite.DxSprite=dxSprite;
			GameStage.DxSprite=dxSprite;

			render.BeginScene(offSurface);

			 device.Clear(ClearFlags.Target, gameStage.BgColor, 1.0f, 0);
			  dxSprite.Begin(SpriteFlags.AlphaBlend);
			   zSorter.Draw(graphics,-9999,Zindex.StageZ);
			   gameStage.Draw(graphics);
			   zSorter.Draw(graphics,Zindex.StageZ,9999,true);
			  dxSprite.End();

			   scoreBoard.Draw(graphics,maryo,hiScore);

			render.EndScene(Filter.None);

#else
			graphics.Clear(gameStage.BgColor);

			 zSorter.Draw(graphics,-9999,Zindex.StageZ);
			 gameStage.Draw(graphics);
			 scoreBoard.Draw(graphics,maryo,hiScore);
			 zSorter.Draw(graphics,Zindex.StageZ,9999,true);

			iLoader.UpdateFrames();
#endif

		  }else
		  if(mode==Waiting){
#if MDX
			if(device==null)return;
			if(device.Disposed)return;

			if(offTexture.Disposed)ResetDevice();

			  scaleMatrix.Scale(1.0f,1.0f,1.0f);
			  dxSprite.Transform=scaleMatrix;
			render.BeginScene(offSurface);
			  dxSprite.Begin(SpriteFlags.AlphaBlend);
			   wScreen.Draw(graphics);
			  dxSprite.End();
			render.EndScene(Filter.None);


#else
			wScreen.Draw(graphics);
#endif
		  }else
		  {
		  }
		 }catch(Exception e){
			DBGn("Exception:EnvManager.Draw:"+e);
		 }
		 drawing=false;
		}

#if MDX
		private Rectangle ctRect = new Rectangle();
		private Rectangle srRect = new Rectangle();
		private Microsoft.DirectX.Matrix scMatrix = new Microsoft.DirectX.Matrix();
		private Microsoft.DirectX.Matrix rtMatrix = new Microsoft.DirectX.Matrix();
		private Vector3 ct = new Vector3();
		private Vector3 ps = new Vector3();
		private double rotate = 0;
		private double zoom = 1;
		private double rad  = Math.PI*2/360;
		private Color full  = Color.FromArgb(255,255,255,255);
		private Color half  = Color.FromArgb(255,128,128,128);
		public void BufferToScreen(Graphics g,int ctx,int cty,int dw,int dh){
			if(device==null)return;
			if(device.Disposed)return;


			double xr = dw / screenWidth;
			double yr = dh / screenHeight;


			
// ₱
			
			  if(mozaic>0){
			   rotate=mozaic*mozaic/3;
//			   xr=xr/(mozaic*.1+1);yr=yr/(mozaic*.1+1);
			   zoom=((mozaic*.1)*(mozaic*.1)+1);
			  }else{
				rotate=0;
				zoom=1;
			  }

			if(zoom != 1.0){
				xr*=zoom;
				yr*=zoom;
			}

			if( (rotate % 360) !=0){
			  double r=rad*rotate;
			double ctxx=(screenWidth  - dw / xr)/2;
			double ctyy=(screenHeight - dh / yr)/2;

			double sw=screenWidth  /xr/2;
			double sh=screenHeight /yr/2;
			  double dx=(sw*Math.Cos(r)-sh*Math.Sin(r));
			  double cxx=ctxx*Math.Cos(r)-ctyy*Math.Sin(r);
			   dx  =  sw -  dx;
			   cxx = ctxx- cxx;
			   ps.X=(int)Math.Round(dx+ctxx);

			  double dy =(sw*Math.Sin(r)+sh*Math.Cos(r));
			  double cyy=ctxx*Math.Sin(r)+ctyy*Math.Cos(r);
			   dy  =  sh -  dy;
			   cyy = ctyy- cyy;
			   ps.Y=(int)Math.Round(dy+ctyy);


				scMatrix.Scale(1.0f,1.0f,1.0f);
				scMatrix.M11=(float)( Math.Cos(r)*xr);
				scMatrix.M12=(float)(-Math.Sin(r)*yr);
				scMatrix.M21=(float)( Math.Sin(r)*xr);
				scMatrix.M22=(float)( Math.Cos(r)*yr);

			}
			else{
				rotate=0;
				scMatrix.Scale(
						(float)xr,
						(float)yr,
						1.0f
						);
				ct.X=0;
				ct.Y=0;
				ps.X=-(int)Math.Round(ctx/xr);
				ps.Y=-(int)Math.Round(cty/yr);
			}

			try{
			dxSprite.Transform=scMatrix;
			device.BeginScene();
			 device.Clear(ClearFlags.Target, Color.Black, 1.0f, 0);
			 dxSprite.Begin(SpriteFlags.None);
//			 dxSprite.Begin(SpriteFlags.AlphaBlend);
			  dxSprite.Draw(
				offTexture,
				screenRect,ps,ct,
				((pause)?half : full)
			  );
			 dxSprite.End();
			device.EndScene();

			device.Present();
			}catch(Exception e){
				DBGn("Exception:EnvManager.BufferToScreen:"+e);
			}
		}
#else
		public void BufferToScreen(Graphics g,int ctx,int cty,int dw,int dh){
			if(clsFlag){
				g.Clear(Color.Black);
				clsFlag=false;
			}
			pixelEffector.Draw(g,ctx,cty,dw,dh);
		}
#endif



		public int GetXYPointObject(double x,double y){
			return gameStage.GetXYPointObject(x,y);
		}
		public int GetXYBlockObject(int x,int y){
			return gameStage.GetXYBlockObject(x,y);
		}
		public int SetXYPointObject(double x,double y,int code,WriteMode flag){
			return gameStage.SetXYPointObject(x,y,code,flag);
		}
		public int SetXYPointObject(double x,double y,int code){
			return gameStage.SetXYPointObject(x,y,code,WriteMode.All);
		}
		public int SetXYBlockObject(int x,int y,int code,WriteMode flag){
			return gameStage.SetXYBlockObject(x,y,code,flag);
		}
		public int SetXYBlockObject(int x,int y,int code){
			return gameStage.SetXYBlockObject(x,y,code,WriteMode.All);
		}
		public int CheckWallBlock(int n){
			return gameStage.CheckWallBlock(n);
		}
		public int CheckFloorBlock(int n){
			return gameStage.CheckFloorBlock(n);
		}
		public int CheckUpperBlock(int n){
			return gameStage.CheckUpperBlock(n);
		}
		public int CheckCoin(int n){
			return gameStage.CheckCoin(n);
		}
		public int CheckEventBlock(int n){
			return gameStage.CheckEventBlock(n);
		}
		public int CheckWaterStream(int n){
			return gameStage.CheckWaterStream(n);
		}

		public string GetBlockImageName(int n){
			return gameStage.GetBlockImageName(n);
		}


		public void DrawStage(Graphics g){gameStage.Draw(g);}
		public bool DrawEntry(Sprite s){return zSorter.Entry(s);}




	//-----------------------------
		public Player GetPlayerObject(){return GetPlayerObject(0);}
		public Player GetPlayerObject(int n){return maryo;}
		public Player SetPlayerObject(Player p){
			maryo=p;
			return maryo;
		}


		public MCharacter GetBoss(){return boss;}

		public void SetBoss(MCharacter m){boss = m;}
		public void KillBoss(){boss = null;}

		public MCharacter GetGoldenAxe(){return goldenAxe;}

		public void SetGoldenAxe(MCharacter m){
			goldenAxe = m;
			//scrollXmax= m.x+m.charSpriteWidth-virtualWidth;
			//if(scrollXmax<scrollXmin)scrollXmax=scrollXmin;
		}
		public void KillGoldenAxe(){
			goldenAxe=null;
			scrollLock=false;
			//scrollXmax=gameStage.cols*blockWidth-virtualWidth;
			enemies.Clear();
		}
	
		public MCharacter GetGoalFlagObject(){return goalFlag;}

		public void SetGoalFlag(double xx,double yy,bool v){
			goalFlag.Set(xx,yy,v,effects);
		}
		public MCharacter GetCastleFlagObject(){
			return castleFlag;
		}
		public void SetCastleFlag(double xx,double yy,bool v){
			castleFlag.Set(xx,yy,v,Effects);
		}
	
		public void SetPrincess(MCharacter m){
			princess = m;
		}
		public MCharacter GetPrincess(){
			return princess;
		}
		public void RescuePrincess(){
			princess=null;
		}
	
	//-----------------------------
		public void Action(){
		  if(pause)return;
			gameClock++;
		  if(mode == Gaming){
			try{
			 if(mozaic>0)mozaic--;
#if MDX
			 if(mozaic>0)mozaic--;
			 else
			 if(mozaic<0)mozaic++;
#endif

			 if(maryo.Animate){
				maryo.SpecialAction();
				Effects.Action();
				return;
			 }

			 maryo.Action();
			 Enemies.Action();
			  Enemies.GroupCollision(maryo);
			 Others.Action();
			  Others.GroupCollision(maryo);
			 Effects.Action();

			 if(maryo.Score>hiScore)hiScore=maryo.Score;


			CallEnemy();


			 if( gameClock % 30 == 0 ){
				if( gameTimer != 0 ){
					gameTimer--;
					if(gameTimer==0){
						maryo.Timeover();
					}
				}
			 }

			 if(maryo.Score>hiScore)hiScore=maryo.Score;


			}catch(GameException e){

			 if(maryo.Score>hiScore)hiScore=maryo.Score;
				DBG("\n");
				switch(e.Code1){
					case GameException.Goal:
						DBGn("GameException:Goal!");
						EventGoalBlock(maryo,e.Code2);
					break;
					case GameException.Die:
						DBGn("GameException:Die!");
						EventDie(maryo);
					break;
					case GameException.Warp:
						DBGn(	"GameException:Warp! ID="+e.Code2+
							" "+lastStage.Count);
						EventWarp(maryo,e.Code2);
					break;
				}
			}
		  }else
		  if(mode == Waiting){
			mode=wScreen.Action();
			mozaic=wScreen.Mozaic;
			
		  }else
		  {
			if(tScreen.Action(graphics)!=Title)GameStart();
		  }
		}


	//-----------------------------
		public void Scroll(){
			 if(scrollLock)return;
			 if(mode==Waiting)return;
			 if(pause)return;

			 if(!maryo.Animate && mode==Gaming){
			  if(scrollLeft>maryo.X  ||  maryo.X > scrollLeft+virtualWidth)
				scrollLeft = maryo.X-virtualWidth  / 2;
			  if(scrollTop >maryo.Y+virtualHeight*.4  ||
			  	  maryo.Y > scrollTop+virtualHeight +blockHeight*2)
				scrollTop  = maryo.Y-virtualHeight / 2;
				
				if(goalFlag!=null){ // 玩XN[~
				  if(CheckScreen(goalFlag)<0){
					scrollLeft += autoScrollX;
					scrollTop  += autoScrollY;
				  }
				  else{
					autoScrollX=0;
					autoScrollY=0;
				  }
				}

			 }
			double pxx=(maryo.X-scrollLeft)/virtualWidth;
			double pyy=(maryo.Y-scrollTop )/virtualHeight;
			
			if(autoScrollX==0){
			 if(pxx>.45 && maryo.XCameraControl > 0 && gameStage.RightScroll)scrollLeft+=maryo.XCameraControl;
			 if(pxx< .3 && maryo.XCameraControl < 0 && gameStage.LeftScroll )scrollLeft+=maryo.XCameraControl;
			}
			if(autoScrollY==0){
			 if(pyy> .5 && maryo.YCameraControl > 0 && gameStage.DownScroll )scrollTop+=maryo.YCameraControl;
			 if(pyy< .4 && maryo.YCameraControl < 0 && gameStage.UpScroll   )scrollTop+=maryo.YCameraControl;
			}
			
			if(scrollLeft < scrollXmin)scrollLeft = scrollXmin;
			if(scrollLeft > scrollXmax)scrollLeft = scrollXmax;

			if(scrollTop  < scrollYmin)scrollTop  = scrollYmin;
			if(scrollTop  > scrollYmax)scrollTop  = scrollYmax;
		}
	//-----------------------------


	//-----------------------------
		public int CheckScreen(MCharacter m){
			return CheckScreen(m.X,m.Y,m.SpriteWidth,m.SpriteHeight);
		}
		public int CheckScreen(Sprite s){
			return CheckScreen(s.X,s.Y,s.Width,s.Height);
		}
		public int CheckScreen(double x,double y,double w,double h){
			 if(x+w < scrollLeft - virtualWidth  *  .25)return -1;
			 if(y+h < scrollTop  - virtualHeight *  .25)return -1;
			 if(x   > scrollLeft + virtualWidth  * 1.25)return -1;
			 if(y   > scrollTop  + virtualHeight * 1.25)return -1;
			
			return 1;
		}












//-----------------------------
	public MCharacter CreateXYBlockObject(int code,double x,double y,int option){
		return CreateObject(code,x*blockWidth,y*blockHeight,option);
	}
	public MCharacter CreateObject(int code,double x,double y,int option){
		MCharacter m= CreateObject(code,x,y,option,Others);
		return m;
	}
	public MCharacter CreateObject(int code,double x,double y,int option,ObjGroup group){
		MCharacter m = dummyObject;
		
		switch(code){
		
			case CharCode.Kinoko: goto case CharCode.Iget;
			case CharCode.Flower: goto case CharCode.Iget;
			case CharCode.Star: goto case CharCode.Iget;
			case CharCode.Tsuta: goto case CharCode.Iget;
			case CharCode.Iget:
			 m=new Item(this,x,y,code,option);
			break;
		
			case CharCode.TenCoin:
			 m=new TenCoinBlock(this,x,y,code,option);
			break;
		
			case CharCode.Bounce:
			 m=new BounceBlock(this,x,y,code,option);
			break;
			case CharCode.DokanUp:goto case CharCode.DokanRight;
			case CharCode.TsutaUp:goto case CharCode.DokanRight;
			case CharCode.DokanDown:goto case CharCode.DokanRight;
			case CharCode.DokanLeft:goto case CharCode.DokanRight;
			case CharCode.DokanRight:
			 m = new WarpHall(this,x,y,code,option);
			break;
		
			case CharCode.JumpBlock:
			 m = new JumpBlock(this,x,y,code,option);
			break;
		
			case CharCode.LiftUp:	goto case CharCode.Kumo_Right;
			case CharCode.LiftDown:	goto case CharCode.Kumo_Right;
			case CharCode.LiftLeft:	goto case CharCode.Kumo_Right;
			case CharCode.LiftRight:goto case CharCode.Kumo_Right;
			case CharCode.LiftFall:	goto case CharCode.Kumo_Right;
			case CharCode.LiftRise:	goto case CharCode.Kumo_Right;
			case CharCode.LiftUpDown:goto case CharCode.Kumo_Right;
			case CharCode.LiftLR:	goto case CharCode.Kumo_Right;
			case CharCode.Kumo_Right:
			 m = new Lift(this,x,y,code,option);
			break;
			case CharCode.BalanceLift:
			 m = new BalanceLift(this,x,y,code,option);
			break;

			case CharCode.FirebarL:	goto case CharCode.FirebarLR;
			case CharCode.FirebarR:	goto case CharCode.FirebarLR;
			case CharCode.FirebarLL:goto case CharCode.FirebarLR;
			case CharCode.FirebarLR:
			 return CreateEnemy(code,x,y,option);
			
			
			case CharCode.GoldenAxe:
			 m = new GoldenAxe(this,x,y,code,option);
			break;

			case CharCode.Castle: goto case CharCode.CastleExcellent;
			case CharCode.CastleExcellent:
			 m=new Castle(this,x,y,code,option);
			break;
			case CharCode.Princess:
			 m=new Princess(this,x,y,code,option);
			break;
		}

		if(m.Mind != Mind.Dummy)m.GroupEntry(group);
		return m;
	}
	//-----------------------------
		public MCharacter CreateXYBlockEnemy(int code,double x,double y,int option){
			return CreateEnemy(code,x*blockWidth,y*blockHeight,option);
		}
		public MCharacter CreateEnemy(int code,double x,double y,int option){
			MCharacter m= CreateEnemy(code,x,y,option,Enemies);
			return m;
		}
		public MCharacter CreateEnemy(int code,double x,double y,int option,ObjGroup group){
			MCharacter m = dummyObject;
				 if(uraMode){
				   if(code==CharCode.Chubo)
						code=CharCode.KameBK;
				 }
			switch(code){
				case CharCode.Chubo:
				 m = new Chubo(this,x,y,code,option);
				break;
				case CharCode.KameGR: goto case CharCode.KameBK;
				case CharCode.KameRD: goto case CharCode.KameBK;
				case CharCode.KameBK:
				 m = new Kame(this,x,y,code,option);
				break;

				case CharCode.Pata:  goto case CharCode.PataH;
				case CharCode.PataV: goto case CharCode.PataH;
				case CharCode.PataH:
				 m = new Pata(this,x,y,code,option);
				break;
				case CharCode.PakkunGR:goto case CharCode.PakkunRD;
				case CharCode.PakkunRD:
				 m = new MonsterFlower(this,x,y,code,option);
				break;

				case CharCode.CannonBall:goto case CharCode.Tamago;
				case CharCode.KamezoBreath:goto case CharCode.Tamago;
				case CharCode.Sakana:goto case CharCode.Tamago;
				case CharCode.Tobiuo:goto case CharCode.Tamago;
				case CharCode.Ika:goto case CharCode.Tamago;
				case CharCode.Tamago:
				 m = new CannonBall(this,x,y,code,option);
				break;
				case CharCode.Toge:
				 m = new Toge(this,x,y,code,option);
				break;

				case CharCode.Cannon:
				 m = new Cannon(this,x,y,code,option);
				break;

				case CharCode.HammerBaki:
				 m = new HammerBaki(this,x,y,code,option);
				break;
				case CharCode.Hammer:
				 m = new Hammer(this,x,y,code,option);
				break;

				case CharCode.Kumozo:
				 m = new Kumozo(this,x,y,code,option);
				break;


				
				case CharCode.FireBubble:
				 m = new FireBubble(this,x,y,code,option);
				break;


				case CharCode.Kamezo:
				 m = new Kamezo(this,x,y,code,option);
				break;
				
				
				case CharCode.FirebarL:goto case CharCode.FirebarLR;
				case CharCode.FirebarR:goto case CharCode.FirebarLR;
				case CharCode.FirebarLL:goto case CharCode.FirebarLR;
				case CharCode.FirebarLR:
				 m = new FireBar(this,x,y,code,option);
				break;

			}
			 if(m!=null && m.Mind!=Mind.Dummy){
			  if(uraMode){
				m.XMove*=2;
			  }
			  m.GroupEntry(group);
			 }
			return m;
		}
	//-----------------------------
		public MCharacter CreateEffect(int code,double x,double y,int option){
			MCharacter m= CreateEffect(code,x,y,option,Effects);
			return m;
		}
		public MCharacter CreateEffect(int code,double x,double y,int option,ObjGroup group){
		 double x2=Math.Floor(x / blockWidth )*blockWidth;
		 double y2=Math.Floor(y / blockHeight)*blockHeight;
			MCharacter m = dummyObject;
			switch(code){
				case CharCode.ScoreEffect:
				 m=new EffectObject(this,x,y,code,option);
				 if(m==null)break;
				 m.Sprite.Text= option.ToString();
				 ((EffectObject)m).ActionCnt=10;
				 m.YMove=-10.1;
				 m.YAccel=1;
				break;
				case CharCode.DyingEffect:
				 m=new DyingCharacter(this,x,y,code,option);
				break;
				case CharCode.BounceCoinEffect:
				 m=new EffectObject(this,x,y,code,option);
				 if(m==null)break;
				 ((EffectObject)m).ActionCnt=10;
				 m.YMove=-30;
				 m.YAccel=4;
				 m.X=x2;
				 m.Y=y2;
				 m.Sprite.BitmapImageName=GetBlockImageName(Blocks.Coin);
				 m.Z=Zindex.StageZ+1;
				break;
				case CharCode.HanabiEffect:
				 m=new EffectObject(this,x,y,code,option);
				 ((EffectObject)m).ActionCnt=10;
				  m.SetSize(32,32);
				  m.Face=GetImageID("img/hanabi.gif");
				break;
				case CharCode.AwaEffect:
				 m=new EffectObject(this,x,y,code,option);
				 ((EffectObject)m).ActionCnt=80;
				 m.YAccel=-.1;
				 m.X=x-8;
				 m.Y=y+8;
				 m.Face=GetImageID("img/awa.gif");
				 m.SetSize(16,16);
				 m.Z=Zindex.StageZ-10;
				break;
				case CharCode.DustEffect:
				 m=CreateMessage("",x,y);
				 m.Sprite.Font=scoreFont;
				 m.Sprite.Color=Color.Yellow;
				break;
				case CharCode.BounceEffect:
				 m=new BounceBlock(this,x,y,code,option);
				break;
				case CharCode.BrokenEffect:
				 m=new BrokenBlock(this,x,y,code,option);
				break;
				case CharCode.ShadowCharacterEffect:
				 m = new ShadowCharacter(this,x,y,code,option);
				break;
			}
			if(m.Mind !=Mind.Dummy)m.GroupEntry(group);
			return m;
		}
		public MCharacter CreateMessage(string str,double x,double y){
			return CreateMessage(x,y,str);
		}
		public MCharacter CreateMessage(double x,double y,String str){
			MCharacter m=CreateEffect(CharCode.ScoreEffect,x,y,0);
			m.Sprite.Text=str;
			return m;
		}

		public MCharacter CenterMessage(string str,double y){
			return CenterMessage(str,y,30*100);
		}
		public MCharacter CenterMessage(string str,double y,int showtime){
			return CenterMessage(str,y,showtime,infoFont);
		}
		public MCharacter CenterMessage(string str,double y,int showtime,
									System.Drawing.Font font){
			double ww=graphics.MeasureString(str,font).Width/wx;
			double hh=font.Height;
			double xx=scrollLeft + ((virtualWidth-ww)/2);
			double yy=scrollTop  + y;//+h*wy;

			EffectObject m=(EffectObject)CreateEffect(CharCode.ScoreEffect,xx,yy,0);
			m.Sprite.Text=str;
			m.Sprite.Font=font;

			m.YMove=0;m.XMove=0;
			m.YAccel=0;
			m.ActionCnt=showtime;
			
			return m;

		}

		private void showInfoText(){
			MCharacter m=CenterMessage(gameStage.InfoText1,virtualHeight*.2);
			 m.XAccel=-.1;
			 m.X=scrollLeft+virtualHeight/2;
			CenterMessage(gameStage.InfoText2,virtualHeight).YAccel=-.1;
		}

	//------------------------------
		//܂
		public void CallEnemy(){
			if(mozaic>0)return;
			if(goalFlag!=null){
				if(CheckScreen(goalFlag)>0)return;
			}
			MCharacter m=null;
			 if(gameStage.Sakanakana)CallEnemy(CharCode.Sakanakana,0,0);
			 if(callEnemyCnt--<=0){
			  if(gameStage.RandomEnemy )m=CallEnemy(-1,0,0);
			  else{
			   if(gameStage.KamezoBreath)m=CallEnemy(CharCode.KamezoBreath,0,0);
			   if(gameStage.Tobiuo      )m=CallEnemy(CharCode.Tobiuo,0,0);
			   if(gameStage.CannonBall  )m=CallEnemy(CharCode.CannonBall,0,0);
			  }
			  if(m!=null)callEnemyCnt= (R.M % callEnemyRate) + 10;
			 }
		}

		public MCharacter CallEnemy(int code){return CallEnemy(code,0,0);}
		public MCharacter CallEnemy(int code,double x,double y){
		 int[] rnd={
			CharCode.KamezoBreath,
			CharCode.CannonBall,
			CharCode.Sakanakana,
			CharCode.Tobiuo,
			CharCode.KamezoBreath,
			CharCode.CannonBall,
			CharCode.Sakanakana,
			CharCode.Tobiuo,
			CharCode.Sakana,
			CharCode.Ika
		 };
		 int[] hf={1,-1,1,-1,1,1,-1,1,-1,-1,1};
		 double lx=scrollLeft;
		 double rx=lx+virtualWidth;
		 ObjGroup E=Enemies;
		 Player p=GetPlayerObject(0);

		 int hface=hf[M(hf.Length)];

		 double px=p.X;
		 double py=p.Y;
		 int cmax=EnemyParameter.LiveCannonMax;
		 if(mode==Title)cmax=30;
		 if(EnemyParameter.LiveCannonCnt>=cmax){
			return null;
		 }
		 switch(code){
			case CharCode.KamezoBreath:
			{
			 double xx=rx+50;
			 double yy=py+(M(4)-2)*blockHeight;
			  if(yy<scrollTop)yy=scrollTop;
			 return CreateEnemy(CharCode.KamezoBreath,xx,yy,0);
			}
			//break;
			
			
			case CharCode.CannonBall:
			{
			 double xx=(hface<0)?(rx+50):(lx-82);
			 double yy=Math.Floor((py+(M(6)-3)*blockHeight)/blockHeight)*blockHeight;
			 return CreateEnemy(CharCode.CannonBall,xx,yy,hface);
			}


			
			case CharCode.Sakanakana:
			{
			 int rr=M(100);
			 MCharacter dd=CallEnemy(( (rr>65) ? CharCode.Ika : CharCode.Sakana ),x,y);
			 dd.ScreenDrop=true;
			 return dd;
			}

			case CharCode.Ika: goto case CharCode.Sakana;
			case CharCode.Sakana:
			{
			  double xx=(hface<0)?(rx):(lx-32);
			  double yy=M(virtualHeight*.8)+scrollTop+virtualHeight*.1;
			  return CreateEnemy(code,xx,yy,hface);
			 }

			case CharCode.Tobiuo:
			{
			 int wx=M(virtualWidth/2);
			 double xx=0;
			 if(hface>0)
			     xx=wx+scrollLeft-virtualWidth*.25;
			 else
			     xx=wx+scrollLeft+virtualWidth*.75;

			 double yy=scrollTop+virtualHeight-3;
			 MCharacter dd=CreateEnemy(CharCode.Tobiuo,xx,yy,hface);
			 return dd;
			}
			
			default:
			{
			 int mm=M(rnd.Length);
			 return CallEnemy(rnd[mm],x,y);
			}
		 }
		} //





















		public void DBG(string str){Console.Write(str);}
		public void DBGn(string str){Console.WriteLine(str);}

		public void Dispose(){
			enemies.Clear();
			others.Clear();
			effects.Clear();

			scoreBoard.Dispose();
			iLoader.Dispose();

#if MDX
			if(dxSprite   != null)dxSprite.Dispose();
			if(render     != null)render.Dispose();
			if(offSurface != null)offSurface.Dispose();
			if(offTexture != null)offTexture.Dispose();
#endif

		}


#if MDX
		public Texture GetTextureData(int n){
			return iLoader.GetTextureData(n);
		}


		public void ResetDevice(Device dev){
			device=dev;
			initMDX(screenRect.Width,screenRect.Height);
		}
		public void ResetDevice(){
			initMDX(screenRect.Width,screenRect.Height);
		}

		private void initMDX(int w,int h){
			if(device==null)return;
			if(device.Disposed)return;
			if(offTexture!=null)offTexture.Dispose();
			offTexture = new Texture(
					device,w,h,
					1,
//					0,
					Usage.RenderTarget,
					Format.X8R8G8B8,
//					Pool.Managed
					Pool.Default
					);

			if(render!=null)render.Dispose();
			render	= new RenderToSurface(device, w ,h, 
					    Format.X8R8G8B8, false, DepthFormat.Unknown);
			if(offSurface!=null)offSurface.Dispose();
			offSurface  = offTexture.GetSurfaceLevel(0);
		
			if(dxSprite!=null)dxSprite.Dispose();
			dxSprite = new Microsoft.DirectX.Direct3D.Sprite(device);
			Sprite.DxSprite=dxSprite;
			GameStage.DxSprite=dxSprite;
//			ScoreBoard.DxSprite=dxSprite;

			scaleMatrix.Scale(1.0f,1.0f,1.0f);
			screenRect.Width=w;
			screenRect.Height=h;
		
			device.Disposing += new System.EventHandler(this.onDeviceDisposing);
			dxSprite.Lost += new EventHandler(this.onSpriteLost);
			render.Lost   += new EventHandler(this.onRenderLost);
			
		}

		private void onSpriteLost(object o,EventArgs e){
		}
		private void onRenderLost(object o,EventArgs e){
		}
		private void onDeviceDisposing(object o,EventArgs e){
			if(render!=null)render.Dispose();
			if(offSurface!=null)offSurface.Dispose();
			if(dxSprite!=null)dxSprite.Dispose();
		}
#endif

	}


	class DummyObject:MCharacter
	{
		public DummyObject(StageManager E,double xx,double yy)
		{
			mind =  Mind.Dummy;
			manager = E;
			x = - 9999; y = - 9999;
		}
	}


}
