namespace mylib
{
	using System.Drawing;
	using System.Drawing.Drawing2D;
	using System.Collections;
	using System.IO;
	using System.Net;

	using System;
	using mylib;
#if MDX
	using System.Drawing.Imaging;
	using Microsoft.DirectX;
	using Microsoft.DirectX.Direct3D;
	using DirectX=Microsoft.DirectX;
	using Direct3D=Microsoft.DirectX.Direct3D;
#endif


	//----------------------------------------------------------------------------
	//----------------------------------------------------------------------------
	//----------------------------------------------------------------------------
	class BlockObjects
	{
		private Image[] im;
		private int[] imID;
		private bool[] isMultiFrame;
		private string[] imName;
		 public Image[] Images{
			get{
				return im;
			}
		 }
		 public Image this[int index]{
			get{
				return im[index];
			}
		 }

		 public string[] FileNames{
			get{
				return imName;
			}
		 }
		 public int Length{
			get{
				return im.Length;
			}
		 }

		private StageManager manager;
		private Image undef;
		private Color bgColor;
		public Color BgColor{
			get{
				return bgColor;
			}
			set{
				bgColor=value;
			}
		}

#if MDX
		private Texture[] textures;
#endif
		
		public BlockObjects(StageManager E)
		{
			im = new Image[256];
			imID = new int[256];
			isMultiFrame = new bool[256];
			imName = new String[256];
			manager = E;
			undef = manager.GetImageData(0);
			bgColor = BGcolor.SkyBlue;
#if MDX
			textures = new Texture[256];
#endif

		}
		public virtual Image GetImage(int n)
		{
			return im[n];
		}
		
		public virtual void  SetA()
		{
			Set(Blocks.SetA);
			bgColor = BGcolor.SkyBlue;
		}
		public virtual void  SetB()
		{
			Set(Blocks.SetB);
			bgColor = BGcolor.UgBlack;
		}
		public virtual void  SetC()
		{
			Set(Blocks.SetC);
			bgColor = BGcolor.WaterBlue;
		}
		public virtual void  SetD()
		{
			Set(Blocks.SetD);
			bgColor = BGcolor.UgBlack;
		}
		public virtual void  SetE()
		{
			//HEAVEN
			Set(Blocks.SetA);
			bgColor = BGcolor.SkyBlue;
			SetBlock(0x01, "img/kumo.gif");
			SetBlock(0x02, "img/tsuta.gif");
			SetBlock(0x04, "img/kumo.gif");
		}
		
		public virtual void Normal()     {SetA();}
		public virtual void Underground(){SetB();}
		public virtual void Water()      {SetC();}
		public virtual void Castle()     {SetD();}
		public virtual void Heaven()     {SetE();}
		
		public virtual void  Set(string[] s)
		{
			int i;
			for (i = 0; i < s.Length; i++)
			{
				if (s[i].Equals("")){
					im[i] = null;
					imID[i] = 0;
					isMultiFrame[i]=false;
					imName[i] = null;
				}
				else
					SetBlock(i, s[i]);
			}
			for (; i < im.Length; i++){
				im[i] = null;
				imID[i] = 0;
				isMultiFrame[i]=false;
				imName[i] = null;
			}
		}
		
		public virtual bool SetBlock(int num, string url)
		{
			int tmp = manager.GetImageID(url);
			im[num] = manager.GetImageData(tmp);
			imID[num] = tmp;
			imName[num] = url;
			isMultiFrame[num]=ILoader.IsMultiFrame(im[num]);

#if MDX
			textures[num] = manager.GetTextureData(tmp);
#endif

			return true;
		}
#if MDX
		public virtual Texture GetTexture(int n)
		{
			return textures[n];
		}
		
		public virtual void UpdateFrames(){
		 for(int i=0;i<im.Length;i++){
			if(isMultiFrame[i]){
				textures[i] = manager.GetTextureData(imID[i]);
			}
		 }
		}

#endif
	}


	//----------------------------------------------------------------------------
	//----------------------------------------------------------------------------
	//----------------------------------------------------------------------------

	public class GameStage{
		private void init()
		{
			warpSrc = new string[30];
			warpStartAnime = new int[30];
			warpStartHFace = new int[30];
			warpStartY = new int[30];
			warpStartX = new int[30];
			vram = new int[Xmax][];
			ram  = new int[Xmax][];
			vdat = new string[Ymax];
			dat  = new string[Ymax];
			for (int i = 0; i < Xmax; i++)
			{
				ram[i] = new int[Ymax];
				vram[i] = new int[Ymax];
			}
			loader = new ArrayList();
		}

		public const int Xmax=300;
		public const int Ymax=300;
		
		public const int WriteVram = 0x01;
		public const int WriteRam  = 0x10;
		public const int WriteAll  = 0x11;
		
		private string[] dat;
		private string[] vdat;
		private int[][] ram;
		private int[][] vram;
		private StageManager manager;

		private BlockObjects blocks;
		private Color bgColor=Color.Blue;
		 public Color BgColor{
			get{
				return bgColor;
			}
		 }
		private int cols=0;
		private int rows=0;
		 public int Cols{
			get{return cols;}
		 }
		 public int Rows{
			get{return rows;}
		 }

		private int stageMode=0;
		private int gameTimer =300;
		private double liftWidth=32*3;
		 public int StageMode{
			get{return stageMode;}
		 }
		 public int GameTimer{
			get{return gameTimer;}
		 }
		 public double LiftWidth{
			get{return liftWidth;}
		 }

		private int plStartX=0;
		private int plStartY=0;
		private int plStartHFace=1;
		private int plStartAnime=0;
		 public int PLStartX{
			get{return plStartX;}
		 }
		 public int PLStartY{
			get{return plStartY;}
		 }
		 public int PLStartHFace{
			get{return plStartHFace;}
		 }
		 public int PLStartAnime{
			get{return plStartAnime;}
		 }

		private int plRestartX=0;
		private int plRestartY=0;
		private int plRestartHFace=1;
		private int plRestartAnime=0;
		 public int PLRestartX{
			get{return plRestartX;}
		 }
		 public int PLRestartY{
			get{return plRestartY;}
		 }
		 public int PLRestartHFace{
			get{return plRestartHFace;}
		 }
		 public int PLRestartAnime{
			get{return plRestartAnime;}
		 }

		private int[] warpStartX;//     = new int[30];
		private int[] warpStartY;//     = new int[30];
		private int[] warpStartHFace;// = new int[30];
		private int[] warpStartAnime;// = new int[30];
		private string[] warpSrc;// = new string[30];
		 public int[] WarpStartX{
			get{return warpStartX;}
		 }
		 public int[] WarpStartY{
			get{return warpStartY;}
		 }
		 public int[] WarpStartHFace{
			get{return warpStartHFace;}
		 }
		 public int[] WarpStartAnime{
			get{return warpStartAnime;}
		 }
		 public string[] WarpSrc{
			get{return warpSrc;}
		 }

		private string infoText1="";
		private string infoText2="";
		 public string InfoText1{
			get{return infoText1;}
		 }
		 public string InfoText2{
			get{return infoText2;}
		 }

		private bool leftScroll  =true;
		private bool rightScroll =true;
		private bool upScroll    =true;
		private bool downScroll  =true;
		 public bool LeftScroll{
			get{return leftScroll;}
		 }
		 public bool RightScroll{
			get{return rightScroll;}
		 }
		 public bool UpScroll{
			get{return upScroll;}
		 }
		 public bool DownScroll{
			get{return downScroll;}
		 }
		private double autoScrollX=0;
		private double autoScrollY=0;
		 public double AutoScrollX{
			get{return autoScrollX;}
		 }
		 public double AutoScrollY{
			get{return autoScrollY;}
		 }


		private bool sakanakana   = false;
		private bool kamezoBreath = false;
		private bool tobiuo       = false;
		private bool cannonBall   = false;
		private bool randomEnemy  = false;
		 public bool Sakanakana{
			get{return sakanakana;}
		 }
		 public bool KamezoBreath{
			get{return kamezoBreath;}
		 }
		 public bool Tobiuo{
			get{return tobiuo;}
		 }
		 public bool CannonBall{
			get{return cannonBall;}
		 }
		 public bool RandomEnemy{
			get{return randomEnemy;}
		 }

		private int maxEnemies = 4;
		private int enemyRate = 100;
		 public int MaxEnemies{
			get{return maxEnemies;}
		 }
		 public int EnemyRate{
			get{return enemyRate;}
		 }


		private double virtualWidth = 512.0;
		private double virtualHeight= 448.0;
		 public double VirtualWidth{
			get{return virtualWidth;}
		 }
		 public double VirtualHeight{
			get{return virtualHeight;}
		 }



		private ArrayList loader=new ArrayList();


#if MDX
		private static Direct3D.Sprite dxSprite;
		 public static Direct3D.Sprite DxSprite{
			set{dxSprite=value;}
		 }
#endif

		public GameStage(StageManager E){
			init();
			manager=E;
			erase();
			blocks=new BlockObjects(manager);

		}
	//-----------------------------
		private void erase(){
			for(int i=0;i<dat.Length;i++){
				dat[i] ="";
			}

			for(int i=0;i<Xmax;i++){
			 for(int j=0;j<Ymax;j++){
				ram[i][j] =0;
				vram[i][j]=0;
			 }
			}

			rows=0;
			cols=0;
		}

	//-----------------------------
		public virtual void CreateDefault(){
		 erase();
		 int i=0;
		dat[i++]= "0000000000000000000000000000000000000000000000000000000000000000";
		dat[i++]= "0000000000000000000000000000000000000000000000000000000000000000";
		dat[i++]= "0000000000000000000000000000000000000000000000000000000000000000";
		dat[i++]= "0000000000000000000000000000000000000000000000000000000000000000";
		dat[i++]= "0000000000000000000006000006000000060000060000000000000000000000";
		dat[i++]= "0000000000000006000006000006000000060000060000060000000000000000";
		dat[i++]= "0000000000000006000006000006000000060000060000060000000000000000";
		dat[i++]= "0000000006000006008080808080808080808080808000060000060000000000";
		dat[i++]= "0000000006000000000101010101010101010101010100000000060000000000";
		dat[i++]= "0000000006000000010001010101010101010101010001000000060000000000";
		dat[i++]= "0000000000000001000001010101010101010101010000010000000000000000";
		dat[i++]= "0006000000000100000001010101010101010101010000000100000000060000";
		dat[i++]= "0006000000010000000001010101010101010101010000000001000000060000";
		dat[i++]= "0006000001000000000001010101010101010101013300000000010000060000";
		dat[i++]= "0e00000100000000000001010101010101010101010000000000000100000000";
		dat[i++]= "0ff3010000000000000000010101010101010101000000000000000001000000";
		dat[i++]= "0000000000000000000000000000000000000000000000000000000000010000";
		dat[i++]= "0100000000000000010180808080808080808080800101000000000000000000";
		dat[i++]= "0100000000000001010000000000000000000000000001010000000000000000";
		dat[i++]= "0100000000000101000000000000000000000000000000010100000000000100";
		dat[i++]= "0100000000010100000000010000000000000001000000000101000000000100";
		dat[i++]= "0100000000010100000000000101010101010000000000000101000000000100";
		dat[i++]= "0100000000000000000000000101010101010101010100000000000000000100";
		dat[i++]= "0021010101010000000000000001010101010000000000000000010000010000";
		dat[i++]= "0000000000010101000000000000010101000001010101010101010000000000";
		dat[i++]= "0000000000000000000606060606060106060606060606010000000100000000";
		dat[i++]= "0000000100000001000606060606060606060606060606000000000001000000";
		dat[i++]= "02020201020202020102020202020202020202020202020102020202010202c8";
		 rows=i;
		 dat[i++]="";
		 Create();
		}

	//-----------------------------
		public virtual void Create(){
		  int i,j;
		  string tmp;
		  int xmax=0;
		 for(i=0;i<dat.Length;i++){
			if(i>=Ymax){manager.DBG("screen Y buffer overflow.\n");break;}
			if(dat[i].Equals(""))break;
			if(dat[i] == null)break;
			tmp=dat[i];
			if(dat[i].Length > Xmax * 2 ){
				manager.DBG("screen X buffer overflow.\n");
				tmp=substring(dat[i],0,Xmax*2);
			}
		  for(j=0;j< tmp.Length ;j+=2){
			ram[j/2][i]  = pInt(substring(tmp,j,j+2),16);
			vram[j/2][i] = ram[j/2][i];
		  }
		  if(xmax<j)xmax=j;
		 }
		 cols=xmax/2;
		 rows=i;
//		 manager.ScrollXmax=cols * manager.BlockWidth  - manager.VirtualWidth;
//		 manager.ScrollYmax=rows * manager.BlockHeight - manager.VirtualHeight;

		} //

		private string substring(string s,int n,int p){
		  try{	return s.Substring(n,p-n);
		  }catch(Exception e){
			manager.DBGn("	'" +s+"': "+e+" ");
			return "00";
		  }
		} //
		private int pInt(string s,int x){
		  try{	return System.Convert.ToInt32(s, x);
		  }catch(Exception e){
			manager.DBGn("	'" +s+"': "+e+" ");
			return 0;
		  }
		} //


		public virtual void CreateObjects(){
		  int i,j;
		  for(j=0;j<cols;j++){
		    for(i=0;i<rows;i++){
			int block=vram[j][i];
			 if(block==Blocks.GoalFlag){
				manager.SetGoalFlag(j*manager.BlockWidth,i*manager.BlockHeight,true);
			 }else
			 if(block>=0xc0){
				manager.SetXYBlockObject(j,i,0, WriteMode.Vram);
				manager.CreateXYBlockObject(block,j,i, 1);
			 }
			 else
			if(block >=0x80){
				manager.SetXYBlockObject(j,i,0, WriteMode.All);
				manager.CreateXYBlockEnemy( block,j,i, 1);
			}
		    }
		  }
		} //

		public virtual void Create(string[] src){
		 Load(src);
		 Create();
		}

	//-----------------------------
		public virtual void Load(string[] src){
		 int i;
		 erase();
		 for(i=0;i<src.Length;i++){
		  if(i>=Xmax)break;
		  dat[i]=substring(src[i],0,Xmax);
		 }
		}
	//-----------------------------
		public virtual int GetXYBlockObject(int x,int y){
			if(y>rows || y<0 || x<0)return 0;
			return ram[x][y];
		}
		public virtual int GetXYPointObject(double xx,double yy){
			int x=(int)Math.Floor(xx/manager.BlockWidth);
			int y=(int)Math.Floor(yy/manager.BlockHeight);
			return GetXYBlockObject(x,y);
		}

	//-----------------------------
		public virtual int SetXYBlockObject(int x,int y,int code,WriteMode option){
			if(y <    0 || x <    0 || y > rows || x > cols)return 0;
			int op=(int)option;

			if(op==0)op=(int)WriteMode.All; // ݊̂
			if( (op & (int)WriteMode.Ram ) != 0)ram[x][y]=code;
			if( (op & (int)WriteMode.Vram) != 0)vram[x][y]=code;
			return code;
		}
		public virtual int SetXYPointObject(double xx,double yy,int code,WriteMode option){
			int x=(int)Math.Floor(xx/manager.BlockWidth);
			int y=(int)Math.Floor(yy/manager.BlockHeight);
			return SetXYBlockObject(x,y,code,option);
		}
		public virtual int CheckWallBlock(int n){
		 if((n<=0) ||
		    (n>=0x50 && n<=0x5f) ||
		    (n>=0x30 && n<=0x3f) ||
		    (n>=0x80))	return -1;
		  switch(n){
			case Blocks.Bridge:
			 return -1;
			case 0x60:
			 return -1;
			case 0x61:
			 return -1;
			case 0x62:
			 return -1;

			case Blocks.Coin:
			 return -2;
			case 0x6d:
			 return -2;
			case 0x6e:
			 return -2;
			case 0x6f:
			 return -2;
			case 0x7f:
			 return -2;

			case 0x6c:
			 return -3;
		  }
		 return 1;
		}
	//-----------------------------
		public virtual int CheckFloorBlock(int n){
		 if((n<=0) ||
		    (n>=0x50 && n<=0x5f) ||
		    (n>=0x30 && n<=0x3f) ||
		    (n>=0x7a && n<=0x7f) ||
		    (n>=0x80))	return -1;
		   if(n>=0x7a && n<=0x7f)return -2;
		  switch(n){
			case Blocks.Coin:
			 return -1;
			case 0x60:
			 return -1;
			case 0x61:
			 return -1;
			case 0x62:
			 return -1;

			case 0x6d:
			 return -2;
			case 0x6e:
			 return -2;
			case 0x6f:
			 return -2;

			case 0x6c:
			 return -1;
		  }
		 return 1;
		}
	//-----------------------------
		public virtual int CheckUpperBlock(int n){
		 if((n<=0) ||
		    (n>=0x50 && n<=0x5f) ||
		    (n>=0x80))	return -1;

		  switch(n){
			case Blocks.Bridge:
			 return -1;
			case Blocks.Coin:
			 return -1;
			case 0x6d:
			 return -1;
			case 0x6e:
			 return -1;
			case 0x6f:
			 return -1;
			case 0x60:
			 return -1;
			case 0x61:
			 return -1;
			case 0x62:
			 return -1;
			case 0x6c:
			 return -1;
		  }
		 return 1;
		}
	//-----------------------------
		public virtual int CheckWaterStream(int n){
			if(n>=0x60 && n<=0x62)return (n-0x5f);
		 return -1;
		}
	//-----------------------------
		public virtual int CheckCoin(int n){
		  switch(n){
			case Blocks.Coin:
			return ItemCode.Coin;
//			return 1;
		  }
		 return -1;
		}
	//-----------------------------
		public virtual int CheckEventBlock(int n){
			if(n>=0x7a &&n<=0x7e)return 1;
			if(n>=0x6e &&n<=0x6f)return 1;
			if(n==0x7f)return 99;
			return -1;
		}


	
	//-----------------------------
#if MDX
	private Vector3 center = new Vector3(0,0,0);
	private Vector3 pos    = new Vector3(0,0,(float)(Zindex.StageZ*.001));
	private DirectX.Matrix scaleMatrix = new DirectX.Matrix();
	private	Rectangle srcRect = new Rectangle(0,0,32,32);
		public virtual void Draw(Graphics g){

			 double w  = manager.BlockWidth  * manager.DrawX; 
			 double h  = manager.BlockHeight * manager.DrawY;
			 double lx = manager.ScrollLeft * manager.DrawX;
			 double ly = manager.ScrollTop  * manager.DrawY;

			int ww=(int)Math.Ceiling(w);
			int hh=(int)Math.Ceiling(h);

			long xx=(long)Math.Floor(manager.ScrollLeft / manager.BlockWidth );
			long yy=(long)Math.Floor(manager.ScrollTop  / manager.BlockHeight);

			long ptX=xx;
			long ptY=yy;

			int blockRight  =(int)Math.Floor(manager.VirtualWidth  / manager.BlockWidth );
			int blockBottom =(int)Math.Floor(manager.VirtualHeight / manager.BlockHeight);

				double lyy=(int)Math.Floor(yy*h-ly);


		pos.X=	(float)(xx/(ww/manager.BlockWidth));
		pos.Y=	(float)(yy/(hh/manager.BlockHeight));

		scaleMatrix.Scale(
			(float)(ww/manager.BlockWidth) ,
			(float)(hh/manager.BlockHeight) ,
			1.0f);
		dxSprite.Transform = scaleMatrix;
//		 dxSprite.Begin(SpriteFlags.AlphaBlend);

		blocks.UpdateFrames();
			for(long j=0;j<=blockBottom;j++){
				long ptXtmp=ptX;

				if(ptY>=0){

				 int fl_lyy=(int)Math.Round(lyy/scaleMatrix.M22);
				 double lxx=xx*w-lx;
				 for(long i=0;i <=blockRight;i++){
				  if(ptXtmp>=0){
					int b=vram[(int)ptXtmp][(int)ptY];
					if(blocks[b] != null){
					 pos.X=(int)Math.Round(lxx/scaleMatrix.M11);
					 pos.Y=fl_lyy;
					dxSprite.Draw(
						blocks.GetTexture(b),
						srcRect,
						center,
						pos,
						Color.FromArgb(255,255,255,255)
					);
					}
				  }
					lxx+=w;
					ptXtmp++;
				 }

				}
				lyy+=h;
				ptY++;
			}

//		dxSprite.End();
			
		}

#else
		public virtual void Draw(Graphics g){

			 double w  = manager.BlockWidth  * manager.DrawX; 
			 double h  = manager.BlockHeight * manager.DrawY;
			 double lx = manager.ScrollLeft * manager.DrawX;
			 double ly = manager.ScrollTop  * manager.DrawY;
			int ww=(int)Math.Ceiling(w);
			int hh=(int)Math.Ceiling(h);

			long xx=(long)Math.Floor(manager.ScrollLeft / manager.BlockWidth );
			long yy=(long)Math.Floor(manager.ScrollTop  / manager.BlockHeight);

			long ptX=xx;
			long ptY=yy;

			int blockRight  =(int)Math.Floor(manager.VirtualWidth  / manager.BlockWidth );
			int blockBottom =(int)Math.Floor(manager.VirtualHeight / manager.BlockHeight);

				double lyy=(int)Math.Floor(yy*h-ly);
			for(long j=0;j<=blockBottom;j++){
				long ptXtmp=ptX;

				if(ptY>=0){

				 int fl_lyy=(int)Math.Floor(lyy);
				 double lxx=xx*w-lx;
				 for(long i=0;i <=blockRight;i++){
				  if(ptXtmp>=0){
					int b=vram[(int)ptXtmp][(int)ptY];
					if(blocks[b] != null){
					 g.DrawImage(blocks[b],
							(int)Math.Floor(lxx),
							fl_lyy,
							ww,
							hh
							);
					}
				  }
					lxx+=w;
					ptXtmp++;
				 }

				}
				lyy+=h;
				ptY++;
			}
			
		}
#endif
	//-----------------------------------------
		public virtual Image GetBlockImage(int num){
			if(num<0 || num>blocks.Length)return null;
			if(num>=0x30 && num <=0x3f)return blocks[Blocks.Empty];
			if(blocks[num]==null)return null;
			return blocks[num];
		}
		public virtual string GetBlockImageName(int num){
			if(num<0 || num>blocks.Length)return "";
			if(num>=0x30 && num <=0x3f)return blocks.FileNames[Blocks.Empty];
			if(blocks[num]==null)return "";
			return  blocks.FileNames[num];
		}




	//-----------------------------------------
		public bool Load(String url){
		 manager.DBG("\n------------------------------\nLoading file "+ url + " ...");
		 erase();
		 ArrayList vc=LoadFrom(url);
		 if( vc.Count > 0 )manager.DBG("OK.\n");
		 else manager.DBG("error!?\n");
		  if( vc.Count < 1 )return false;

		 manager.DBG("loading param...");
		 Hashtable mp=new Hashtable();
		 int i=0;
		  for (IEnumerator e = vc.GetEnumerator(); e.MoveNext(); ){
			string cc = (string)e.Current;
//		  foreach(string cc in (string)vc){

			if(cc[0]=='#'){
				continue;
			}
			if(cc[0]=='"'){
				cc=cc.Replace(' ','0').Replace('"',' ').Replace(';',' ').Replace(',',' ').Replace('\r',' ').Trim();
				if(i<Ymax)dat[i++]=cc;
				manager.DBG(".");
			}else{
			  if(cc.IndexOf("=")>0){
				int c=cc.IndexOf("=");
				manager.DBG(".");
				string cname  = ((substring(cc,0,c)  ).ToUpper()).Trim();
				string cvalue =  (cc.Substring(c+1)).Replace('\r',' ').Replace('\n',' ').Trim();
				if(mp.Contains(cname)){
					manager.DBGn("\nIvV"+cname+ "͓dݒ̉\܂B");
					mp.Remove(cname);
				}
				mp.Add(cname,cvalue);
	//			manager.DBG("cname + "+ cname+" = cvalue= " +cvalue);
			  }
			}
		 }
		 if(i==0){
			dat[13]="01010101010101010101010101";
			dat[14]="01010101010101010101010101";
		 }
		manager.DBG("OK\n");

		manager.DBG("get param...");
		  getParam(mp);
		manager.DBG("OK\n");
		  
		manager.DBG("create stage...");
		  Create();
		manager.DBG("OK\n");

		 mp=null;
		 return true;
		}

		protected virtual void getParam(Hashtable mp){
		  gameTimer =param(mp,"timer",GameConfig.DefaultGameTimer);
		
		  
		  plStartX   = param(mp,"PLStartX",0);
		  plStartY   = param(mp,"PLStartY",0);
		  plStartHFace=param(mp,"PLStartHFace",1);
		  plStartAnime=param(mp,"PLStartAnime",0);

		  plRestartX   = param(mp,"PLRestartX",plStartX);
		  plRestartY   = param(mp,"PLRestartY",plStartY);
		  plRestartHFace=param(mp,"PLRestartHFace",plStartHFace);
		  plRestartAnime=param(mp,"PLRestartAnime",plStartAnime);

		  leftScroll = param(mp,"scrollLeft" , true);
		  rightScroll= param(mp,"scrollRight", true);
		  upScroll   = param(mp,"scrollUp"   ,false);
		  downScroll = param(mp,"scrollDown" ,false);
		  autoScrollX= param(mp,"autoScrollX",0.0);
		  autoScrollY= param(mp,"autoScrollY",0.0);

		  virtualWidth  = param(mp,"virtualWidth" ,SystemConfig.DefaultVScreenWidth);
		  virtualHeight = param(mp,"virtualHeight",SystemConfig.DefaultVScreenHeight);

		  sakanakana   = param(mp,"Sakanakana",false);
		  kamezoBreath = param(mp,"KamezoBreath",false);
		  tobiuo       = param(mp,"Tobiuo",false);
		  cannonBall   = param(mp,"CannonBall",false);
		  randomEnemy  = param(mp,"randomEnemy",false);
		  maxEnemies   = param(mp,"maxEnemmies",EnemyParameter.LiveCannonMax); 
		  enemyRate    = param(mp,"enemyRate",50); 


		  infoText1  = param(mp,"infoText1" ,"");
		  infoText2  = param(mp,"infoText2" ,"");
		  for(int i=0;i<warpStartX.Length;i++){
			warpStartX[i]     = param(mp,"warpStartX"+i    ,0);
			warpStartY[i]     = param(mp,"warpStartY"+i    ,0);
			warpStartHFace[i] = param(mp,"warpStartHFace"+i,1);
			warpStartAnime[i] = param(mp,"warpStartAnime"+i,0);
			warpSrc[i]	  = param(mp,"warpTo"+i,"");
			warpSrc[i]	  = param(mp,"warpTo0X"+System.Convert.ToString(i, 16),warpSrc[i]);
		  }

		  stageMode  = param(mp,"stageMode",0);
		  liftWidth  = param(mp,"liftWidth",manager.BlockWidth*3);
		  switch(stageMode){

			case mylib.StageMode.Normal:
			 blocks.Normal();
			break;
			case mylib.StageMode.Underground:
			 blocks.Underground();
			break;
			case mylib.StageMode.Water:
			 blocks.Water();
			break;
			case mylib.StageMode.Castle:
			 blocks.Castle();
			break;
			case mylib.StageMode.Heaven:
			 blocks.Heaven();
			break;

			default:
			 blocks.Normal();
			break;
		  }

		  string cl=param(mp,"bgColor" ,"");
		  if(!cl.Equals(""))
			bgColor= Color.FromArgb(
					System.Convert.ToInt32(substring(cl,0,2),16),
					System.Convert.ToInt32(substring(cl,2,4),16),
					System.Convert.ToInt32(substring(cl,4,6),16)
					);
		  else	bgColor=blocks.BgColor;

		  for(int i=0;i<256;i++){
			string str=param(mp,"Block"+i    ,"");
			 str=param(mp,"Block0x"+System.Convert.ToString(i,16)    ,str);
			if(!str.Equals("")){
				blocks.SetBlock(i,str);
	//			manager.DBGn("str="+str);
				str="";
			}

		  }
		}

		private int param(Hashtable mp,string k,int def){string key=k.ToUpper();
			try{
			 if(mp.ContainsKey(key)){
				return System.Int32.Parse((string)mp[key]);
			 }
			 else	return def;
			}catch(Exception e){
				manager.DBGn(e+" value = "+key+" : " + mp[key]);
				return def;
			}
		}
		private bool param(Hashtable mp,string k,bool def){string key=k.ToUpper();
			try{
			 if(mp.ContainsKey(key))
				return System.Boolean.Parse((string)mp[key]);
			 else	return def;
			}catch(Exception e){
				manager.DBG(e+" value = "+key+" : " + mp[key]);
				return def;
			}
		}
		private double param(Hashtable mp,string k,double def){string key=k.ToUpper();
			try{
			 if(mp.ContainsKey(key))
				return System.Double.Parse((string)mp[key]);
			 else	return def;
			}catch(Exception e){
				manager.DBG(e+" value = "+key+" : " + mp[key]);
				return def;
			}
		}
		private string param(Hashtable mp,string k,string def){string key=k.ToUpper();
			 if(mp.ContainsKey(key))return (string)mp[key];
			 else	return def;
		}


		protected virtual ArrayList LoadFrom(string str){
		  loader.Clear();
			try{

			if(manager.CodeBase.IndexOf("http://")==0){
				return LoadFromURI(str);
			}

				StreamReader input = 
				  new StreamReader(
				      str,System.Text.Encoding.UTF8
				    );

			string line;
			 while( (line = input.ReadLine()) != null) {
				if(!line.Equals(""))loader.Add(line);
			 }

			// foreach(object cc in loader){ }

			}catch(Exception e){
				manager.DBGn("\n error"+ e);
			}
		 return loader;
		}

		protected virtual ArrayList LoadFromURI(string str){
		  loader.Clear();
			try{
			Uri uri;			
			  if( str.IndexOf("http://")<0 )uri =new Uri(manager.CodeBase + str);
			  else uri=new Uri(str);

				StreamReader input = 
				  new StreamReader(
				      WebRequest.Create(uri).GetResponse().GetResponseStream(),
				      System.Text.Encoding.UTF8
				    );

			string line;
			 while( (line = input.ReadLine()) != null) {
				if(!line.Equals(""))loader.Add(line);
			 }
			}catch(Exception e){
				manager.DBGn("\n error"+ e);
			}
		 return loader;
		}



	} // end GameStage
}