namespace mylib{
	using System.Drawing;
	using System;
	using mylib;
#if MDX
	using System.Drawing.Imaging;
	using Microsoft.DirectX;
	using Microsoft.DirectX.Direct3D;
	using DirectX=Microsoft.DirectX;
	using Direct3D=Microsoft.DirectX.Direct3D;
#endif

	class ScoreBoard{

		private StageManager manager;
		
		private System.Drawing.Font  font;
		private Brush brush;
		private Brush brushY = Brushes.Yellow;
		private Brush brushR = Brushes.Red;

		private float fontHeight;
		
		private float worldWidth;
		private float maryoWidth;
		private float coinWidth;
		private float hiscWidth;
		private float scoreWidth;
		private float timeWidth;

		private Image kinta;
		private float kintaWidth;

#if MDX
		private Direct3D.Sprite dxSprite;
		private Direct3D.Font dxFont;
		private Point point = new Point(0,0);
		private Texture kintaTX;


		private Color color  = Color.White;
		private Color colorY = Color.Yellow;
		private Color colorR = Color.Red;
#endif

		public ScoreBoard(StageManager m){
			manager=m;
		}


		public ScoreBoard(StageManager m,Graphics g,System.Drawing.Font f,Brush b){
			manager=m;
			Init(g,f,b);

		}


		public void Init(Graphics g,System.Drawing.Font f,Brush b){
			font=f;
			brush=b;
			fontHeight=font.Height;
			worldWidth=getWidth(g,"WORLD");
			maryoWidth=getWidth(g,"MARYO");
			coinWidth =getWidth(g,"COIN");
			hiscWidth =getWidth(g,"HI-SCORE");
			scoreWidth=getWidth(g,"SCORE");
			timeWidth =getWidth(g,"TIME");
			
			kintaWidth=getWidth(g,"0")+5;

			kinta=manager.GetImage("img/coinA.gif");

#if MDX
			dxFont   = new Direct3D.Font(manager.DxDevice,font);
			dxSprite = new Direct3D.Sprite(manager.DxDevice);
			kintaTX  = manager.GetTextureData(manager.GetImageID("img/coinA.gif"));
#endif
		}
	
		private float getWidth(Graphics g,string s){
			return g.MeasureString(s,font).Width;
		}
	
	
		public void Draw(Graphics g,Player p,int hiScore){
#if MDX
			Color tmpColor=color;
			dxSprite.Begin(SpriteFlags.AlphaBlend);
			 if(p.Fire)tmpColor=colorY;
#else
			Brush tmpBrush=brush;
			 if(p.Fire)tmpBrush=brushY;
#endif


			int world = manager.GetWorldNumber();
			int area  = manager.GetAreaNumber();
			int vw=(int)Math.Round(manager.ScreenWidth*.98);

			float baseX=0;
			float h=fontHeight;
			float ws=0;
			float wn=0;



			ws = worldWidth;
			wn = getWidth(g,world.ToString() + "-"+ area);
			baseX=vw/10-ws/2;
#if MDX
			 drawText("WORLD",baseX,0,tmpColor);
			 drawText(world + "-"+ area,baseX+(ws-wn)/2,h,tmpColor);
#else
			 g.DrawString("WORLD",font,tmpBrush,baseX,0);
			 g.DrawString(""+world + "-"+ area,font,tmpBrush,baseX+(ws-wn)/2,h);
#endif



			ws = maryoWidth;
			wn = getWidth(g,p.Life.ToString());
			baseX=vw*1/5+vw/10-ws/2;
#if MDX
			 drawText("MARYO",baseX,0,tmpColor);
			 drawText(p.Life.ToString(),baseX+(ws-wn),h,tmpColor);
#else
			 g.DrawString("MARYO",font,tmpBrush,baseX,0);
			 g.DrawString(p.Life.ToString(),font,tmpBrush,baseX+(ws-wn),h);
#endif




			ws = coinWidth;
			wn = getWidth(g,""+p.Coin.ToString());
			baseX =vw*2/5-ws/2+vw/10-ws/2;
#if MDX
			 drawText("COIN",baseX,0,tmpColor);
			 drawText(p.Coin.ToString(),baseX+(ws-wn),h,tmpColor);
			  drawKinta(baseX-kintaWidth/3,h,kintaWidth,h);
#else
			 g.DrawString("COIN",font,tmpBrush,baseX,0);
			 g.DrawString(p.Coin.ToString(),font,tmpBrush,baseX+(ws-wn),h);
			  g.DrawImage(kinta,baseX,h,kintaWidth,h);
#endif




			ws = hiscWidth;
			wn = scoreWidth;
			baseX=vw*4/5-ws/2;
#if MDX
			 drawText("HI-SCORE",baseX,0,tmpColor);
			 drawText("SCORE",baseX+(ws-wn),h,tmpColor);
#else
			 g.DrawString("HI-SCORE",font,tmpBrush,baseX,0);
			 g.DrawString("SCORE",font,tmpBrush,baseX+(ws-wn),h);
#endif



			baseX=vw;
			ws = getWidth(g,hiScore.ToString());
			wn = getWidth(g,p.Score.ToString());
#if MDX
			 drawText(hiScore.ToString(),baseX-ws,0,tmpColor);
			 drawText(p.Score.ToString(),baseX-wn,h,tmpColor);
#else
			 g.DrawString(hiScore.ToString(),font,tmpBrush, baseX-ws,0);
			 g.DrawString(p.Score.ToString(),font,tmpBrush, baseX-wn,h);
#endif



			ws = timeWidth;
			wn = getWidth(g,manager.GameTimer.ToString());
			baseX=vw*3/5-ws + vw/10-ws/2;
#if MDX
			if(manager.GameTimer<100)tmpColor=colorR;
			 drawText("TIME",baseX,0,tmpColor);
			 drawText(manager.GameTimer.ToString(),baseX+(ws-wn),h,tmpColor);
#else
			if(manager.GameTimer<100)tmpBrush=brushR;
			 g.DrawString("TIME",font,tmpBrush,baseX,0);
			 g.DrawString(manager.GameTimer.ToString(),font,tmpBrush,baseX+(ws-wn),h);
#endif




#if MDX
			dxSprite.End();
#endif
		}//



		public void Dispose(){
#if MDX
			if(dxFont  !=null)dxFont.Dispose();
			if(dxSprite!=null)dxSprite.Dispose();
#endif
		}

#if MDX
		private int r(double b){
				return (int)Math.Round(b);
		}
		private void setP(double x,double y){
			point.X=r(x);
			point.Y=r(y);
		}
		private void drawText(string str,double x,double y,Color col){
			setP(x,y);
			dxFont.DrawText(dxSprite,str,point,col);
		}
		
	// Ȃʓ|	
		private Rectangle srcRect = new Rectangle();
		private Microsoft.DirectX.Matrix scaleMatrix = new Microsoft.DirectX.Matrix();
		private Vector3 center = new Vector3();
		private Vector3 pos = new Vector3();
		private void drawKinta(double x,double y,double w,double h){
			double xr=w/kinta.Width;
			double yr=h/kinta.Height;

			srcRect.Width=kinta.Width;
			srcRect.Height=kinta.Height;
			pos.X=(int)Math.Round(x/xr);
			pos.Y=(int)Math.Round(y/yr);
			pos.Z=1;
			
			scaleMatrix.Scale(
					(float)(xr),
					(float)(yr),
					0f);
			dxSprite.Transform= scaleMatrix;

			dxSprite.Draw(
					kintaTX,
					srcRect,
					center,
					pos,
					Color.FromArgb(255,255,255,255)
			);

			scaleMatrix.Scale(1.0f,1.0f,0f);
			dxSprite.Transform= scaleMatrix;
		}
#endif
	}
}